# Copyright (C) 2014 Collabora Ltd
#
# Author: Sjoerd Simons <sjoerd.simons@collabora.co.uk>
#
# This file is part of Linaro Image Tools.
#
# Linaro Image Tools is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# Linaro Image Tools is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Linaro Image Tools.  If not, see <http://www.gnu.org/licenses/>.


SECTOR_SIZE = 512  # bytes
# align on 4 MiB
PART_ALIGN_S = 4 * 1024 * 1024 / SECTOR_SIZE

from parted import PARTITION_BOOT, PARTITION_BIOS_GRUB
import yaml
from mako.template import Template

flagmap = { "boot": PARTITION_BOOT,
            "bios_grub": PARTITION_BIOS_GRUB }

def align_up(value, align):
    """Round value to the next multiple of align."""
    return (value + align - 1) / align * align

def align_partition(min_start, min_length, start_alignment, end_alignment):
    """Compute partition start and end offsets based on specified constraints.

    :param min_start: Minimal start offset of partition
    :param min_lengh: Minimal length of partition
    :param start_alignment: Alignment of this partition
    :param end_alignment: Alignment of the data following this partition
    :return: start offset, end offset (inclusive), length
    """
    start = align_up(min_start, start_alignment)
    # end offset is inclusive, so substact one
    end = align_up(start + min_length, end_alignment) - 1
    # and add one to length
    length = end - start + 1
    return start, end, length

class Partition(object):
    """ partition data object """
    def __init__(self, name, start, end, number,
            ptype = None, filesystem = None):
        self.name = name
        self.start = start
        self.end = end
        self.number = number
        self.filesystem = filesystem
        self.ptype = ptype
        # Map of subvolumes of a given partition, only used if the partition
        # is btrfs
        self.subvolumes = {}
        self.bootable = False
        self.flags = []
        self.device = None
        self.uuid = None

    def set_bootable(self):
        self.bootable = True

    def add_subvolume (self, name, default = False):
        assert name not in self.subvolumes.keys()
        if default:
            assert True not in self.subvolumes.values()
        self.subvolumes[name] = default

    def set_device(self, device):
        self.device = device

    def set_uuid(self, uuid):
        self.uuid = uuid

class Mountpoint(object):
    def __init__(self, partition, mountpoint, subvol = None, options = None):
        self.partition = partition
        self.mountpoint = mountpoint
        self.subvol = subvol
        self.options = options

class PartitionLayout(object):
    """ Partition layout information object """
    def __init__(self, board, layout = None):
        self.partitions = []
        self.partition_type = None
        self.gpt_gap = None

        v = {}
        v['loader_start'], v['loader_end'], v['loader_len'] = align_partition(
            board.LOADER_START_S, board.LOADER_MIN_SIZE_S, SECTOR_SIZE, PART_ALIGN_S)

        v['boot_start'], v['boot_end'], v['boot_len'] = align_partition(
            v['loader_end'] + 1, board.BOOT_MIN_SIZE_S, PART_ALIGN_S, PART_ALIGN_S)

        v['root_start'], v['_root_end'], v['root_len'] = align_partition(
            v['boot_end'] + 1, board.ROOT_MIN_SIZE_S, PART_ALIGN_S, PART_ALIGN_S)

        if layout == None:
            layout = board.partition_layout
            if not layout.startswith("/"):
              layout = '/'.join(('', 'etc',
                    'linaro-image-tools', 'partition-layouts', layout))
              layout = board.get_file("partition_layout", default=layout)

        print layout

        if (self.is_template (layout)):
            self.load_template(layout,v )
        else:
            self.load_yaml(layout, v)

    def get_next_line (self, f):
        while True:
            line = f.readline()
            if line == None or not line.startswith('#'):
                    break
        return line.rstrip()

    def is_template (self, layout):
        with open(layout) as f:
            while True:
                line = self.get_next_line (f)
                if line == None:
                    break
                return line.split(' ')[0] in [ "mklabel", "mktable" ]
        return False

    def get_boot(self):
        return self.boot_partition

    def get_root_layout(self):
        return self.root_layout

    def find_partition(self, name):
        for p in self.partitions:
            if p.name == name:
                return p
        return None

    def find_partition_number(self, number):
        print "looking for %d" % (number)
        for p in self.partitions:
            if p.number == number:
                return p
        return None

    def get_next_partition_number (self, parttype):
        number = len(self.partitions) + 1
        if parttype == "logical":
          number = 5
          for pp in self.partitions:
             if pp.ptype == "logical":
                number += 1
        return number

    def load_yaml(self, layout, v):
        y = None
        with open(layout) as f:
            y = yaml.safe_load(f)
        self.partition_type = y['type']
        self.gpt_gap = y.get('gpt_gap')
        for v in y.get('partitions'):
            parttype = v.get('type')
            number = self.get_next_partition_number (parttype)
            p = Partition (v['name'], v['start'], v['end'], number,
                ptype = parttype, filesystem = v.get('fs'))
            self.partitions.append(p)
            for s in v.get('subvolumes', []):
                p.add_subvolume(s['name'], s.get('default', False))
            flags = v.get('flags', [])
            for f in flags:
                if f in flagmap.keys():
                    p.flags.append(flagmap[f])
                else:
                    print "Unknown partition flag: %s" % f
        boot = y['boot']
        self.boot_partition = self.find_partition(boot['partition'])
        assert self.boot_partition != None
        root = y['root']
        r = self.find_partition(root['partition'])
        assert r != None
        m = Mountpoint(r, "/",
                subvol = root.get('subvol', None),
                options = root.get('options', None))
        self.root_layout = [ m ];

        # Append other filesystems as required
        for f in root.get('mountpoints', []):
            p = self.find_partition(f['partition'])
            assert p != None
            m = Mountpoint(p, f['mountpoint'],
                subvol = f.get('subvol', None),
                options = f.get('options', None))
            self.root_layout.append(m)

    def parse_template_line (self, line):
        if len(line) < 5 or line[0] == '#':
            return
        command = line.split()
        if self.partition_type == None:
            assert command[0] in [ 'mklabel', 'mktable' ]
            self.partition_type = command[1]
            if self.partition_type == 'gpt' and len(command) > 2:
                self.gpt_gap = command[2]
            return

        if command[0] == 'mkpart':
            fs = None
            t = None
            if len(command) == 4:
                _d, parttype, start, end = command
            else:
                _d, parttype, fs, start, end = command

            number = self.get_next_partition_number (parttype)
            if self.partition_type == 'gpt':
                name = parttype
            else:
                name = "part %d" % number
                t = parttype

            p = Partition(name, start, end, number,
                ptype = t, filesystem = fs)
            self.partitions.append(p)

        if command[0] == 'set':
            assert command[3] == 'on'
            p = self.find_partition_number (int(command[1]))
            if command[2] in flagmap.keys():
                flag = flagmap[command[2]]
                p.flags.append(flag)
            else:
                print "Unknown partition flag: %s" % command[2]

    def load_template(self, layout, v):
        partitionTemplate = Template(filename=layout)
        lines = unicode(partitionTemplate.render(**v)).split("\n")
        is_gpt = False
        for l in lines:
            self.parse_template_line (l)
        # Legacy l-i-t logic is that the first partition with the bootable
        # flag is boot and the next "normal" partition is root
        for p in self.partitions:
            if PARTITION_BOOT in p.flags:
                self.boot_partition = p
        assert self.boot_partition
        p = self.find_partition_number (self.boot_partition.number + 1)
        m = Mountpoint(p, "/")
        self.root_layout = [ m ];
